/*
    Simplified user-agent stylesheet for HTML4 in tests.
*/
html, body, div, h1, h2, h3, h4, ol, p, ul, hr, pre, section
                { display: block; }
li              { display: list-item }
head            { display: none }
pre             { white-space: pre }
br:before       { content: '\A'; white-space: pre-line }
ol              { list-style-type: decimal; counter-reset: list-item }

table           { display: table;
                  box-sizing: border-box }
tr              { display: table-row }
thead           { display: table-header-group }
tbody           { display: table-row-group }
tfoot           { display: table-footer-group }
col             { display: table-column }
colgroup        { display: table-column-group }
td, th          { display: table-cell }
caption         { display: table-caption }
